//=============================================================================
// PrimitiveDemo.cpp by Kyoung Shin Park (Spring 2010)
//
// Demonstrates primitive's types.
//
// Controls: Use mouse to orbit and zoom; use the 'W' and 'S' keys to 
//           alter the height of the camera.
//=============================================================================

#include "d3dApp.h"
#include "DirectInput.h"
#include <crtdbg.h>
#include "GfxStats.h"
#include <list>
#include "Vertex.h"

const VertexPos LineListVertices[] = 
{
    {-1.0f,  0.0f, 2.0f }, // Line #1
    { 0.0f,  1.0f, 2.0f },

    { 0.5f,  1.0f, 2.0f }, // Line #2
    { 0.5f, -1.0f, 2.0f },

	{ 1.0f, -0.5f, 2.0f }, // Line #3
    {-1.0f, -0.5f, 2.0f },
};

const VertexPos LineStripVertices[] = 
{
    { 0.5f, 0.5f, 2.0f },
    { 1.0f, 0.0f, 2.0f },
    { 0.0f,-1.0f, 2.0f },
    {-1.0f, 0.0f, 2.0f },
	{ 0.0f, 0.0f, 2.0f },
    { 0.0f, 1.0f, 2.0f },
};

const VertexPos TriangleListVertices[] = 
{	{-1.0f,  0.0f, 2.0f},	// triangle #1
	{ 0.0f,  1.0f, 2.0f},
	{ 1.0f,  0.0f, 2.0f},
	{-1.0f, -1.0f, 2.0f},	// triangle #2
	{ 0.0f,  0.0f, 2.0f},
	{ 1.0f, -1.0f, 2.0f}
};

const VertexPos TriangleStripVertices[] = 
{
    {-2.0f, 0.0f, 2.0f },
    {-1.0f, 1.0f, 2.0f },
    {-1.0f, 0.0f, 2.0f },
    { 0.0f, 1.0f, 2.0f },
    { 0.0f, 0.0f, 2.0f },
    { 1.0f, 1.0f, 2.0f },
	{ 1.0f, 0.0f, 2.0f },
	{ 2.0f, 1.0f, 2.0f }
};

const VertexPos TriangleFanVertices[] = 
{
    { 0.0f,-1.0f, 2.0f },
    {-1.0f, 0.0f, 2.0f },
    {-0.5f, 0.5f, 2.0f },
    { 0.0f, 1.0f, 2.0f },
    { 0.5f, 0.5f, 2.0f },
    { 1.0f, 0.0f, 2.0f }
};

const VertexPos CubeVertices[8] =
{
	{-1.0f, -1.0f, -1.0f},
	{-1.0f,  1.0f, -1.0f},
	{ 1.0f,  1.0f, -1.0f},
	{ 1.0f, -1.0f, -1.0f},
	{-1.0f, -1.0f,  1.0f},
	{-1.0f,  1.0f,  1.0f},
	{ 1.0f,  1.0f,  1.0f},
	{ 1.0f, -1.0f,  1.0f}
};

const WORD CubeIndices[36] =
{
	// front side
	0, 1, 2,
	0, 2, 3,

	// back side
	4, 6, 5,
	4, 7, 6,

	// left side
	4, 5, 1,
	4, 1, 0,

	// right side
	3, 2, 6,
	3, 6, 7,

	// top
	1, 5, 6,
	1, 6, 2,

	// bottom
	4, 0, 3,
	4, 3, 7
};

class PrimitiveDemo : public D3DApp
{
public:
	PrimitiveDemo(HINSTANCE hInstance, std::string winCaption, D3DDEVTYPE devType, DWORD requestedVP);
	~PrimitiveDemo();

	bool checkDeviceCaps();
	void onLostDevice();
	void onResetDevice();
	void updateScene(float dt);
	void drawScene();

	// Helper methods
	void buildCubeVertexIndexBuffer();
	void buildPrimitiveVertexBuffer();
	void buildViewMtx();
	void buildProjMtx();

private:
	GfxStats* mGfxStats;
	
	IDirect3DVertexBuffer9* m_CubeVB;
	IDirect3DIndexBuffer9*  m_CubeIB;
	IDirect3DVertexBuffer9* m_LineListVB;
	IDirect3DVertexBuffer9* m_LineStripVB;
	IDirect3DVertexBuffer9* m_TriangleListVB;
	IDirect3DVertexBuffer9* m_TriangleStripVB;
	IDirect3DVertexBuffer9* m_TriangleFanVB;

	float mCameraRotationY;
	float mCameraRadius;
	float mCameraHeight;

	D3DXMATRIX mView;
	D3DXMATRIX mProj;

	D3DPRIMITIVETYPE mPrimitiveMode; 
};


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE prevInstance,
				   PSTR cmdLine, int showCmd)
{
	// Enable run-time memory check for debug builds.
	#if defined(DEBUG) | defined(_DEBUG)
		_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
	#endif

	PrimitiveDemo app(hInstance, "Primitive Demo", D3DDEVTYPE_HAL, D3DCREATE_HARDWARE_VERTEXPROCESSING);
	gd3dApp = &app;

	DirectInput di(DISCL_NONEXCLUSIVE|DISCL_FOREGROUND, DISCL_NONEXCLUSIVE|DISCL_FOREGROUND);
	gDInput = &di;

	return gd3dApp->run();
}

PrimitiveDemo::PrimitiveDemo(HINSTANCE hInstance, std::string winCaption, D3DDEVTYPE devType, DWORD requestedVP)
: D3DApp(hInstance, winCaption, devType, requestedVP)
{
	if(!checkDeviceCaps())
	{
		MessageBox(0, "checkDeviceCaps() Failed", 0, 0);
		PostQuitMessage(0);
	}

	mGfxStats = new GfxStats();

	mCameraRadius    = 10.0f;
	mCameraRotationY = 1.2 * D3DX_PI;
	mCameraHeight    = 5.0f;

	buildCubeVertexIndexBuffer();
	buildPrimitiveVertexBuffer();
	mPrimitiveMode = D3DPT_POINTLIST;

	onResetDevice();

	InitAllVertexDeclarations();
}

PrimitiveDemo::~PrimitiveDemo()
{
	delete mGfxStats;
	ReleaseCOM(m_CubeVB);
	ReleaseCOM(m_CubeIB);
	ReleaseCOM(m_LineListVB);
	ReleaseCOM(m_LineStripVB);
	ReleaseCOM(m_TriangleListVB);
	ReleaseCOM(m_TriangleStripVB);
	ReleaseCOM(m_TriangleFanVB);
	DestroyAllVertexDeclarations();
}

bool PrimitiveDemo::checkDeviceCaps()
{
	return true;
}

void PrimitiveDemo::onLostDevice()
{
	mGfxStats->onLostDevice();
}

void PrimitiveDemo::onResetDevice()
{
	mGfxStats->onResetDevice();

	// The aspect ratio depends on the backbuffer dimensions, which can 
	// possibly change after a reset.  So rebuild the projection matrix.
	buildProjMtx();
}

void PrimitiveDemo::updateScene(float dt)
{
	// One cube has 8 vertice and 12 triangles.
	mGfxStats->setVertexCount(8);
	mGfxStats->setTriCount(12);
	mGfxStats->update(dt);

	// Get snapshot of input devices.
	gDInput->poll();

	if( gDInput->keyDown(DIK_SPACE) )	{ 
		if (mPrimitiveMode == D3DPT_POINTLIST)
			mPrimitiveMode = D3DPT_LINELIST;
		else if (mPrimitiveMode == D3DPT_LINELIST)
			mPrimitiveMode = D3DPT_LINESTRIP;
		else if (mPrimitiveMode == D3DPT_LINESTRIP)
			mPrimitiveMode = D3DPT_TRIANGLELIST;
		else if (mPrimitiveMode == D3DPT_TRIANGLELIST)
			mPrimitiveMode = D3DPT_TRIANGLESTRIP;
		else if (mPrimitiveMode == D3DPT_TRIANGLESTRIP)
			mPrimitiveMode = D3DPT_TRIANGLEFAN;
		else if (mPrimitiveMode == D3DPT_TRIANGLEFAN)
			mPrimitiveMode = D3DPT_POINTLIST;
	}

	// Check input.
	if( gDInput->keyDown(DIK_W) )	 
		mCameraHeight   += 25.0f * dt;
	if( gDInput->keyDown(DIK_S) )	 
		mCameraHeight   -= 25.0f * dt;

	// Divide by 50 to make mouse less sensitive. 
	mCameraRotationY += gDInput->mouseDX() / 100.0f;
	mCameraRadius    += gDInput->mouseDY() / 25.0f;

	// If we rotate over 360 degrees, just roll back to 0
	if( fabsf(mCameraRotationY) >= 2.0f * D3DX_PI ) 
		mCameraRotationY = 0.0f;

	// Don't let radius get too small.
	if( mCameraRadius < 5.0f )
		mCameraRadius = 5.0f;

	// The camera position/orientation relative to world space can 
	// change every frame based on input, so we need to rebuild the
	// view matrix every frame with the latest changes.
	buildViewMtx();
}


void PrimitiveDemo::drawScene()
{
	// Clear the backbuffer and depth buffer.
	HR(gd3dDevice->Clear(0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0xffffffff, 1.0f, 0));

	// World matrix is identity.
	D3DXMATRIX W;
	D3DXMatrixIdentity(&W);
	HR(gd3dDevice->SetTransform(D3DTS_WORLD, &W));
	HR(gd3dDevice->SetTransform(D3DTS_VIEW, &mView));
	HR(gd3dDevice->SetTransform(D3DTS_PROJECTION, &mProj));

	HR(gd3dDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME));
	HR(gd3dDevice->BeginScene());

	// draw different primitive type
	switch (mPrimitiveMode)
	{
		case D3DPT_POINTLIST:		// 1 (cube demo)
			gd3dDevice->SetStreamSource(0, m_CubeVB, 0, sizeof(VertexPos));
			gd3dDevice->SetIndices(m_CubeIB);
			gd3dDevice->SetVertexDeclaration(VertexPos::Decl);
			gd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, 8, 0, 12);
			break;
		case D3DPT_LINELIST:		// 2
			gd3dDevice->SetStreamSource(0, m_LineListVB, 0, sizeof(VertexPos));
			gd3dDevice->SetVertexDeclaration(VertexPos::Decl);
			gd3dDevice->DrawPrimitive(D3DPT_LINELIST, 0, 3);
			break;
		case D3DPT_LINESTRIP:		// 3
			gd3dDevice->SetStreamSource(0, m_LineStripVB, 0, sizeof(VertexPos));
			gd3dDevice->SetVertexDeclaration(VertexPos::Decl);
			gd3dDevice->DrawPrimitive(D3DPT_LINESTRIP, 0, 5);
			break;
		case D3DPT_TRIANGLELIST:	// 4
			gd3dDevice->SetStreamSource(0, m_TriangleListVB, 0, sizeof(VertexPos));
			gd3dDevice->SetVertexDeclaration(VertexPos::Decl);
			gd3dDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 2);
			break;
		case D3DPT_TRIANGLESTRIP:	// 5
			gd3dDevice->SetStreamSource(0, m_TriangleStripVB, 0, sizeof(VertexPos));
			gd3dDevice->SetVertexDeclaration(VertexPos::Decl);
			gd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 6);
			break;
		case D3DPT_TRIANGLEFAN:		// 6
			gd3dDevice->SetStreamSource(0, m_TriangleFanVB, 0, sizeof(VertexPos));
			gd3dDevice->SetVertexDeclaration(VertexPos::Decl);
			gd3dDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, 0, 4);
			break;
		default:
			break;
	}

	mGfxStats->display();

	HR(gd3dDevice->EndScene());

	// Present the backbuffer.
	HR(gd3dDevice->Present(0, 0, 0, 0));
}

void PrimitiveDemo::buildCubeVertexIndexBuffer()
{
	// Obtain a pointer to a new vertex buffer.
	HR(gd3dDevice->CreateVertexBuffer(8 * sizeof(VertexPos), D3DUSAGE_WRITEONLY,
		0, D3DPOOL_MANAGED, &m_CubeVB, 0));

	// Now lock it to obtain a pointer to its internal data, and write the
	// cube's vertex data.
	VertexPos* v = 0;
	HR(m_CubeVB->Lock(0, 0, (void**)&v, 0));
	for (int i = 0; i < 8; i++) {
		v[i] = VertexPos(CubeVertices[i]);
	}
	HR(m_CubeVB->Unlock());

	// Obtain a pointer to a new index buffer.
	HR(gd3dDevice->CreateIndexBuffer(36 * sizeof(WORD), D3DUSAGE_WRITEONLY,
		D3DFMT_INDEX16, D3DPOOL_MANAGED, &m_CubeIB, 0));

	// Now lock it to obtain a pointer to its internal data, and write the
	// cube's index data.
	WORD* k = 0;
	HR(m_CubeIB->Lock(0, 0, (void**)&k, 0));
	memcpy(k, CubeIndices, sizeof(CubeIndices));
	HR(m_CubeIB->Unlock());
}

void PrimitiveDemo::buildPrimitiveVertexBuffer()
{
	VertexPos* v = NULL;

	// Create the vertex buffer for LINELIST.
	// Fill the buffers with the linelist data.
	gd3dDevice->CreateVertexBuffer(
		6 * sizeof(VertexPos),	// size in bytes
		D3DUSAGE_WRITEONLY,		// flags
		0,						// vertex format
		D3DPOOL_MANAGED,		// managed memory pool
		&m_LineListVB,			// return create vertex buffer
		0);					// not used - set to 0

	v = NULL;
	m_LineListVB->Lock(0, 0, (void**)&v, 0);
	for (int i = 0; i < 6; i++) {
		v[i] = VertexPos(LineListVertices[i]);
	}
	m_LineListVB->Unlock();

	// Create the vertex buffer for LINESTRIP.
	// Fill the buffers with the linestrip data.
	gd3dDevice->CreateVertexBuffer(
		6 * sizeof(VertexPos), // size in bytes
		D3DUSAGE_WRITEONLY, // flags
		0,        // vertex format
		D3DPOOL_MANAGED,    // managed memory pool
		&m_LineStripVB,    // return create vertex buffer
		0);                 // not used - set to 0

	v = NULL;
	m_LineStripVB->Lock(0, 0, (void**)&v, 0);
	for (int i = 0; i < 6; i++) {
		v[i] = VertexPos(LineStripVertices[i]);
	}
	m_LineStripVB->Unlock();

	// Create the vertex buffer for TRIANGLELIST.
	// Fill the buffers with the trianglelist data.
	gd3dDevice->CreateVertexBuffer(
		6 * sizeof(VertexPos), // size in bytes
		D3DUSAGE_WRITEONLY, // flags
		0,        // vertex format
		D3DPOOL_MANAGED,    // managed memory pool
		&m_TriangleListVB,    // return create vertex buffer
		0);                 // not used - set to 0

	v = NULL;
	m_TriangleListVB->Lock(0, 0, (void**)&v, 0);
	for (int i = 0; i < 6; i++) {
		v[i] = VertexPos(TriangleListVertices[i]);
	}
	m_TriangleListVB->Unlock();

	// Create the vertex buffer for TRIANGLESTRIP.
	// Fill the buffers with the trianglestrip data.
	gd3dDevice->CreateVertexBuffer(
		8 * sizeof(VertexPos), // size in bytes
		D3DUSAGE_WRITEONLY, // flags
		0,        // vertex format
		D3DPOOL_MANAGED,    // managed memory pool
		&m_TriangleStripVB,    // return create vertex buffer
		0);                 // not used - set to 0

	v = NULL;
	m_TriangleStripVB->Lock(0, 0, (void**)&v, 0);
	for (int i = 0; i < 8; i++) {
		v[i] = VertexPos(TriangleStripVertices[i]);
	}
	m_TriangleStripVB->Unlock();

	// Create the vertex buffer for TRIANGLEFAN.
	// Fill the buffers with the trianglefan data.
	gd3dDevice->CreateVertexBuffer(
		6 * sizeof(VertexPos), // size in bytes
		D3DUSAGE_WRITEONLY, // flags
		0,        // vertex format
		D3DPOOL_MANAGED,    // managed memory pool
		&m_TriangleFanVB,    // return create vertex buffer
		0);                 // not used - set to 0

	v = NULL;
	m_TriangleFanVB->Lock(0, 0, (void**)&v, 0);
	for (int i = 0; i < 6; i++) {
		v[i] = VertexPos(TriangleFanVertices[i]);
	}
	m_TriangleFanVB->Unlock();
}

void PrimitiveDemo::buildViewMtx()
{
	float x = mCameraRadius * cosf(mCameraRotationY);
	float z = mCameraRadius * sinf(mCameraRotationY);
	D3DXVECTOR3 pos(x, mCameraHeight, z);
	D3DXVECTOR3 target(0.0f, 0.0f, 0.0f);
	D3DXVECTOR3 up(0.0f, 1.0f, 0.0f);
	D3DXMatrixLookAtLH(&mView, &pos, &target, &up);
}

void PrimitiveDemo::buildProjMtx()
{
	float w = (float)md3dPP.BackBufferWidth;
	float h = (float)md3dPP.BackBufferHeight;
	D3DXMatrixPerspectiveFovLH(&mProj, D3DX_PI * 0.25f, w/h, 1.0f, 5000.0f);
}